import os
import traceback
from pathlib import Path
from typing import Any, Dict

import bittensor as bt
from transformers import AutoModelForCausalLM, AutoTokenizer

from taoverse.model.data import Model, ModelId
from taoverse.model.storage.disk import utils
from taoverse.model.storage.local_model_store import LocalModelStore


class DiskModelStore(LocalModelStore):
    """Local storage based implementation for storing and retrieving a model on disk."""

    def __init__(self, base_dir: str):
        self.base_dir = base_dir
        os.makedirs(utils.get_local_miners_dir(base_dir), exist_ok=True)

    def get_path(self, hotkey: str) -> str:
        """Returns the path to where this store would locate this hotkey."""
        return utils.get_local_miner_dir(self.base_dir, hotkey)

    def store_model(self, hotkey: str, model: Model) -> ModelId:
        """Stores a trained model locally."""

        save_directory = utils.get_local_model_snapshot_dir(
            self.base_dir, hotkey, model.id
        )

        # Note that the revision argument here does not affect the directory path like with hugging face downloads.
        model.pt_model.save_pretrained(
            save_directory=save_directory,
            revision=model.id.commit,
            safe_serialization=True,
        )

        # If the model has a tokenizer also save that.
        if model.tokenizer is not None:
            model.tokenizer.save_pretrained(
                save_directory=save_directory,
                revision=model.id.commit,
                safe_serialization=True,
            )

        # Return the same model id used as we do not edit the commit information.
        return model.id

    def retrieve_model(
        self, hotkey: str, model_id: ModelId, kwargs: Dict[str, Any]
    ) -> Model:
        """Retrieves a trained model locally."""

        pretrained_model_name_or_path = utils.get_local_model_snapshot_dir(
            self.base_dir, hotkey, model_id
        )

        model = AutoModelForCausalLM.from_pretrained(
            pretrained_model_name_or_path=pretrained_model_name_or_path,
            revision=model_id.commit,
            local_files_only=True,
            use_safetensors=True,
            **kwargs,
        )

        # Always try to retrieve a tokenizer from the model directory. If we do not find one leave it None on the Model.
        tokenizer = None
        try:
            # Do not use the kwargs for the model load here. If needed in the future a separate kwargs can be plumbed.
            tokenizer = AutoTokenizer.from_pretrained(
                pretrained_model_name_or_path=pretrained_model_name_or_path,
                revision=model_id.commit,
                local_files_only=True,
                use_safetensors=True,
            )
        except Exception:
            pass

        return Model(id=model_id, pt_model=model, tokenizer=tokenizer)

    def delete_unreferenced_models(
        self, valid_models_by_hotkey: Dict[str, ModelId], grace_period_seconds: int
    ):
        """Check across all of local storage and delete unreferenced models out of grace period."""
        # Expected directory structure is as follows.
        # self.base_dir/models/hotkey/models--namespace--name/snapshots/commit/config.json + other files.

        # Create a set of valid model paths up to where we expect to see the actual files.
        valid_model_paths = set()
        for hotkey, model_id in valid_models_by_hotkey.items():
            valid_model_paths.add(
                utils.get_local_model_snapshot_dir(self.base_dir, hotkey, model_id)
            )

        # For each hotkey path on disk using listdir to go one level deep.
        miners_dir = Path(utils.get_local_miners_dir(self.base_dir))
        hotkey_subfolder_names = [d.name for d in miners_dir.iterdir() if d.is_dir()]

        for hotkey in hotkey_subfolder_names:
            try:
                # Reconstruct the path from the hotkey
                hotkey_path = utils.get_local_miner_dir(self.base_dir, hotkey)

                # If it is not in valid_hotkeys and out of grace period remove it.
                if hotkey not in valid_models_by_hotkey:
                    deleted_hotkey = utils.remove_dir_out_of_grace(
                        hotkey_path, grace_period_seconds
                    )
                    if deleted_hotkey:
                        bt.logging.trace(
                            f"Removed directory for unreferenced hotkey: {hotkey}."
                        )

                else:
                    # Check all the models--namespace--name subfolder paths.
                    hotkey_dir = Path(hotkey_path)
                    model_subfolder_paths = [
                        str(d) for d in hotkey_dir.iterdir() if d.is_dir()
                    ]

                    # Check all the snapshots subfolder paths
                    for model_path in model_subfolder_paths:
                        model_dir = Path(model_path)
                        snapshot_subfolder_paths = [
                            str(d) for d in model_dir.iterdir() if d.is_dir()
                        ]

                        # Check all the commit paths.
                        for snapshot_path in snapshot_subfolder_paths:
                            snapshot_dir = Path(snapshot_path)
                            commit_subfolder_paths = [
                                str(d) for d in snapshot_dir.iterdir() if d.is_dir()
                            ]

                            # Reached the end. Check all the actual commit subfolders for the files.
                            for commit_path in commit_subfolder_paths:
                                if commit_path not in valid_model_paths:
                                    deleted_model = utils.remove_dir_out_of_grace(
                                        commit_path, grace_period_seconds
                                    )
                                    if deleted_model:
                                        bt.logging.trace(
                                            f"Removing directory for unreferenced model at: {commit_path}."
                                        )
            except Exception:
                # Catch the exception so we continue with the rest of the cleanup.
                bt.logging.warning(traceback.format_exc())



import base64
import datetime
import hashlib
import os
import shutil
import sys

from taoverse.model.data import ModelId


def get_local_miners_dir(base_dir: str) -> str:
    return os.path.join(base_dir, "models")


def get_local_miner_dir(base_dir: str, hotkey: str) -> str:
    return os.path.join(get_local_miners_dir(base_dir), hotkey)


# Hugging face stores models under models--namespace--name/snapshots/commit when downloading.
def get_local_model_dir(base_dir: str, hotkey: str, model_id: ModelId) -> str:
    return os.path.join(
        get_local_miner_dir(base_dir, hotkey),
        "models" + "--" + model_id.namespace + "--" + model_id.name,
    )


def get_local_model_snapshot_dir(base_dir: str, hotkey: str, model_id: ModelId) -> str:
    return os.path.join(
        get_local_model_dir(base_dir, hotkey, model_id),
        "snapshots",
        model_id.commit,
    )


def get_hf_download_path(local_path: str, model_id: ModelId) -> str:
    return os.path.join(
        local_path,
        "models" + "--" + model_id.namespace + "--" + model_id.name,
        "snapshots",
        model_id.commit,
    )


def get_newest_datetime_under_path(path: str) -> datetime.datetime:
    newest_filetime = 0

    # Check to see if any file at any level was modified more recently than the current one.
    for cur_path, dirnames, filenames in os.walk(path):
        for filename in filenames:
            path = os.path.join(cur_path, filename)
            try:
                mod_time = os.stat(path).st_mtime
                if mod_time > newest_filetime:
                    newest_filetime = mod_time
            except:
                pass

    if newest_filetime == 0:
        return datetime.datetime.min

    return datetime.datetime.fromtimestamp(newest_filetime)


def remove_dir_out_of_grace_by_datetime(
    path: str, grace_period_seconds: int, last_modified: datetime.datetime
) -> bool:
    """Removes a dir if the last modified time is out of grace period secs. Returns if it was deleted."""
    grace = datetime.timedelta(seconds=grace_period_seconds)

    if last_modified < datetime.datetime.now() - grace:
        shutil.rmtree(path=path, ignore_errors=True)
        return True

    return False


def remove_dir_out_of_grace(path: str, grace_period_seconds: int) -> bool:
    """Removes a dir if the last modified time is out of grace period secs. Returns if it was deleted."""
    last_modified = get_newest_datetime_under_path(path)
    return remove_dir_out_of_grace_by_datetime(
        path, grace_period_seconds, last_modified
    )


def realize_symlinks_in_directory(path: str) -> int:
    """Realizes all symlinks in the given directory, moving the linked file to the location. Returns count removed."""
    realized_symlinks = 0

    for cur_path, dirnames, filenames in os.walk(path):
        for filename in filenames:
            path = os.path.abspath(os.path.join(cur_path, filename))
            # Get path resolving symlinks if encountered
            real_path = os.path.realpath(path)
            # If different then move
            if path != real_path:
                realized_symlinks += 1
                shutil.move(real_path, path)

    return realized_symlinks


def get_hash_of_file(path: str) -> str:
    blocksize = 64 * 1024
    file_hash = hashlib.sha256()
    with open(path, "rb") as fp:
        while True:
            data = fp.read(blocksize)
            if not data:
                break
            file_hash.update(data)
    return base64.b64encode(file_hash.digest()).decode("utf-8")


def get_hash_of_directory(path: str) -> str:
    dir_hash = hashlib.sha256()

    # Recursively walk everything under the directory for files.
    for cur_path, dirnames, filenames in os.walk(path):
        # Ensure we walk future directories in a consistent order.
        dirnames.sort()
        # Ensure we walk files in a consistent order.
        for filename in sorted(filenames):
            path = os.path.join(cur_path, filename)
            file_hash = get_hash_of_file(path)
            dir_hash.update(file_hash.encode())

    return base64.b64encode(dir_hash.digest()).decode("utf-8")



import datetime
import os
import shutil
import time
import unittest
from tempfile import TemporaryDirectory

import taoverse.model.storage.disk.utils as utils
from taoverse.model.data import ModelId


class TestDiskUtils(unittest.TestCase):
    def setUp(self):
        self.sep = os.path.sep

    def test_get_local_miners_dir(self):
        with TemporaryDirectory() as base_dir:
            miners_dir = utils.get_local_miners_dir(base_dir)

            expected_path = base_dir + self.sep + "models"
            self.assertEqual(miners_dir, expected_path)

    def test_get_local_miner_dir(self):
        hotkey = "test-hotkey"

        with TemporaryDirectory() as base_dir:
            miner_dir = utils.get_local_miner_dir(base_dir, hotkey)

            expected_path = base_dir + self.sep + "models" + self.sep + hotkey
            self.assertEqual(miner_dir, expected_path)

    def test_get_local_model_dir(self):
        hotkey = "test-hotkey"
        namespace = "test-namespace"
        name = "test-name"
        commit = "test-commit"
        model_id = ModelId(
            namespace=namespace,
            name=name,
            hash="test-hash",
            commit=commit,
            competition_id=1,
        )

        with TemporaryDirectory() as base_dir:
            model_dir = utils.get_local_model_dir(base_dir, hotkey, model_id)

            expected_path = (
                base_dir
                + self.sep
                + "models"
                + self.sep
                + hotkey
                + self.sep
                + "models--"
                + namespace
                + "--"
                + name
            )
            self.assertEqual(model_dir, expected_path)

    def test_get_local_model_snapshot_dir(self):
        hotkey = "test-hotkey"
        namespace = "test-namespace"
        name = "test-name"
        commit = "test-commit"
        model_id = ModelId(
            namespace=namespace,
            name=name,
            hash="test-hash",
            commit=commit,
            competition_id=1,
        )

        with TemporaryDirectory() as base_dir:
            model_dir = utils.get_local_model_snapshot_dir(base_dir, hotkey, model_id)

            expected_path = (
                base_dir
                + self.sep
                + "models"
                + self.sep
                + hotkey
                + self.sep
                + "models--"
                + namespace
                + "--"
                + name
                + self.sep
                + "snapshots"
                + self.sep
                + commit
            )
            self.assertEqual(model_dir, expected_path)

    def test_get_hf_download_path_dir(self):
        hotkey = "test-hotkey"
        namespace = "test-namespace"
        name = "test-name"
        commit = "test-commit"
        model_id = ModelId(
            namespace=namespace,
            name=name,
            hash="test-hash",
            commit=commit,
            competition_id=1,
        )

        with TemporaryDirectory() as base_dir:
            hf_download_path_dir = utils.get_hf_download_path(
                utils.get_local_miner_dir(base_dir, hotkey), model_id
            )

            expected_path = (
                base_dir
                + self.sep
                + "models"
                + self.sep
                + hotkey
                + self.sep
                + "models--"
                + namespace
                + "--"
                + name
                + self.sep
                + "snapshots"
                + self.sep
                + commit
            )
            self.assertEqual(hf_download_path_dir, expected_path)

    def test_get_newest_datetime_under_path(self):
        # Write two files to ensure we get the time of the most recently written one.
        file_name1 = "test1.txt"
        file_name2 = "test2.txt"
        with TemporaryDirectory() as base_dir:
            path1 = base_dir + os.path.sep + file_name1
            path2 = base_dir + os.path.sep + file_name2

            file1 = open(path1, "w")
            file1.write("test text 1.")
            file1.close()

            time.sleep(1)

            file2 = open(path2, "w")
            file2.write("test text 2.")
            file2.close()

            last_modified_file1 = datetime.datetime.fromtimestamp(
                os.path.getmtime(path1)
            )
            last_modified_file2 = datetime.datetime.fromtimestamp(
                os.path.getmtime(path2)
            )
            last_modified_dir = utils.get_newest_datetime_under_path(base_dir)

            # Confirm that the 2nd file was modified later and that this matches the newest in the directory.
            self.assertLess(last_modified_file1, last_modified_file2)
            self.assertEqual(last_modified_dir, last_modified_file2)

    def test_get_newest_datetime_under_path_empty(self):
        last_modified_expected = datetime.datetime.min

        with TemporaryDirectory() as base_dir:
            last_modified_actual = utils.get_newest_datetime_under_path(base_dir)

            self.assertEqual(last_modified_actual, last_modified_expected)

    def test_remove_dir_out_of_grace(self):
        file_name = "test.txt"
        with TemporaryDirectory() as base_dir:
            path = base_dir + self.sep + file_name

            file = open(path, "w")
            file.write("test text.")
            file.close()

            # Sleep to ensure we are out of grace.
            time.sleep(1)

            self.assertTrue(os.path.exists(base_dir))
            deleted = utils.remove_dir_out_of_grace(base_dir, 0)
            self.assertTrue(deleted)
            self.assertFalse(os.path.exists(base_dir))

    def test_remove_dir_out_of_grace_in_grace(self):
        file_name = "test.txt"
        with TemporaryDirectory() as base_dir:
            path = base_dir + self.sep + file_name

            file = open(path, "w")
            file.write("test text.")
            file.close()

            self.assertTrue(os.path.exists(base_dir))
            deleted = utils.remove_dir_out_of_grace(base_dir, 60)
            self.assertFalse(deleted)
            self.assertTrue(os.path.exists(base_dir))

    def test_get_hash_of_file(self):
        file_name = "test.txt"
        with TemporaryDirectory() as base_dir:
            path = base_dir + self.sep + file_name

            file = open(path, "w")
            file.write("test text.")
            file.close()

            # Obtained by running openssl dgst -sha256 -binary test.txt | base64
            expected_file_hash = "tXNDvHVzqYIRiUx0rvK+M5+Lu4OLzhfPJH+gf7HvCeA="
            actual_file_hash = utils.get_hash_of_file(path)

            self.assertEqual(actual_file_hash, expected_file_hash)

    def test_get_hash_of_directory(self):
        with TemporaryDirectory() as base_dir:
            # Make two sub directories.
            dir_1 = base_dir + self.sep + "dir1"
            dir_2 = base_dir + self.sep + "dir2"

            # Write the same two files to both sub directories.
            file_name_1 = "test1.txt"
            file_name_2 = "test2.txt"
            path_1_file_1 = dir_1 + os.path.sep + file_name_1
            path_1_file_2 = dir_1 + os.path.sep + file_name_2
            path_2_file_1 = dir_2 + os.path.sep + file_name_1
            path_2_file_2 = dir_2 + os.path.sep + file_name_2

            path_2_file_2 = dir_2 + os.path.sep + file_name_2
            file_paths = [path_1_file_1, path_1_file_2, path_2_file_1, path_2_file_2]

            os.mkdir(dir_1)
            os.mkdir(dir_2)

            for file_path in file_paths:
                file = open(file_path, "w")
                file.write("test text.")
                file.close()

            # Test that both sub directories have an equal hash.
            dir_1_hash = utils.get_hash_of_directory(dir_1)
            dir_2_hash = utils.get_hash_of_directory(dir_2)
            self.assertEqual(dir_1_hash, dir_2_hash)

            # Test that the hash for the overall directory does not equal the sub directory.
            base_dir_hash = utils.get_hash_of_directory(base_dir)
            self.assertNotEqual(base_dir_hash, dir_1_hash)

    def test_realize_symlinks_in_directory(self):
        # We can't use temporary directories because all files in the directoy are symlinks.
        # Instead, we create our own non-temporary directory, that's deleted at the end of the test.
        base_dir = "test-symlinks"
        shutil.rmtree(path=base_dir, ignore_errors=True)
        os.mkdir(base_dir)

        end_file_dir = base_dir + self.sep + "end_files"
        symlink_source_dir = base_dir + self.sep + "symlink"

        regular_file = end_file_dir + self.sep + "test_file.txt"
        symlink_source = symlink_source_dir + self.sep + "symlink_source.txt"
        symlink_dest = end_file_dir + self.sep + "symlink_end.txt"

        # Make a regular file
        os.mkdir(end_file_dir)
        file = open(regular_file, "w")
        file.write("test text.")
        file.close()

        # Make a symlinked file
        os.mkdir(symlink_source_dir)
        file = open(symlink_source, "w")
        file.write("symlink source test text.")
        file.close()

        os.symlink(os.path.abspath(symlink_source), os.path.abspath(symlink_dest))

        # Confirm we see 3 files
        pre_file_count = 0
        for _, _, files in os.walk(base_dir):
            pre_file_count += len(files)
        self.assertEqual(pre_file_count, 3)

        realized_files = utils.realize_symlinks_in_directory(end_file_dir)

        # Confirm 1 file got realized and there are two total now.
        self.assertEqual(realized_files, 1)

        post_file_count = 0
        for _, _, files in os.walk(base_dir):
            post_file_count += len(files)
        self.assertEqual(post_file_count, 2)

        shutil.rmtree(path=base_dir, ignore_errors=True)


if __name__ == "__main__":
    unittest.main()



from collections import defaultdict, deque
from typing import List, Optional

from taoverse.model.data import ModelId, ModelMetadata
from taoverse.model.storage.model_metadata_store import ModelMetadataStore


class FakeModelMetadataStore(ModelMetadataStore):
    """Fake implementation for storing and retrieving metadata about a model."""

    def __init__(self):
        self.current_block = 1
        self.metadata = dict()
        self.store_errors = deque()
        self.injected_metadata = defaultdict(deque)

    async def store_model_metadata(self, hotkey: str, model_id: ModelId):
        """Fake stores model metadata for a specific hotkey."""

        # Return an injected error if we have one.
        if len(self.store_errors) > 0:
            raise self.store_errors.popleft()

        model_metadata = ModelMetadata(id=model_id, block=self.current_block)
        self.current_block += 1

        self.metadata[hotkey] = model_metadata

    async def store_model_metadata_exact(
        self, hotkey: str, model_metadata: ModelMetadata
    ):
        """Fake stores model metadata for a specific hotkey."""

        # Return an injected error if we have one.
        if len(self.store_errors) > 0:
            raise self.store_errors.popleft()

        self.metadata[hotkey] = model_metadata

    async def retrieve_model_metadata(self, hotkey: str) -> Optional[ModelMetadata]:
        """Retrieves model metadata for a specific hotkey"""

        if len(self.injected_metadata[hotkey]) > 0:
            return self.injected_metadata[hotkey].popleft()

        return self.metadata[hotkey] if hotkey in self.metadata else None

    def inject_model_metadata(self, hotkey: str, metadata: ModelMetadata):
        """Injects a metadata for hotkey to be returned instead of the expected response."""
        self.injected_metadata[hotkey].append(metadata)

    def inject_store_errors(self, errors: List[Exception]):
        """Injects a list of errors to be raised on the next N calls to store_model_metadata."""
        self.store_errors.extend(errors)

    def reset(self):
        """Resets the store to its initial state."""
        self.current_block = 1
        self.metadata = dict()
        self.store_errors = deque()



import asyncio
import unittest
from typing import Dict, List, Tuple

import torch
from transformers import (
    BartForCausalLM,
    FalconForCausalLM,
    GemmaForCausalLM,
    GPT2Config,
    GPT2LMHeadModel,
    GPTNeoXForCausalLM,
    LlamaForCausalLM,
    MistralForCausalLM,
    PhiForCausalLM,
)

from taoverse.model import utils
from taoverse.model.competition.data import (
    Competition,
    ModelConstraints,
    NormValidationConstraints,
)
from taoverse.model.competition.epsilon import FixedEpsilon
from taoverse.model.data import Model, ModelId, ModelMetadata
from taoverse.model.model_tracker import ModelTracker
from taoverse.model.model_updater import MinerMisconfiguredError, ModelUpdater
from taoverse.model.storage.disk.disk_model_store import DiskModelStore
from tests.model.storage.fake_model_metadata_store import FakeModelMetadataStore
from tests.model.storage.fake_remote_model_store import FakeRemoteModelStore
from tests.utils import get_test_model


class TestModelUpdater(unittest.TestCase):
    COMPETITION_ID = 1

    MODEL_CONSTRAINTS_BY_COMPETITION_ID: Dict[int, ModelConstraints] = {
        COMPETITION_ID: ModelConstraints(
            max_model_parameter_size=6_900_000_000,
            sequence_length=4096,
            allowed_architectures=[
                MistralForCausalLM,
                LlamaForCausalLM,
                BartForCausalLM,
                FalconForCausalLM,
                GPTNeoXForCausalLM,
                PhiForCausalLM,
                GemmaForCausalLM,
            ],
            tokenizer="Xenova/gpt-4",
            kwargs={
                "torch_dtype": torch.bfloat16,
            },
            eval_block_delay=1200,  # ~4 hours.
            norm_validation_constraints=NormValidationConstraints(
                norm_eps_soft=200,
                norm_eps_soft_percent_threshold=0.15,
                norm_eps_hard=1000,
            ),
            epsilon_func=FixedEpsilon(0.005),
            max_bytes=15*1024*1024*1024,  # 15 GB
        ),
    }

    COMPETITION_SCHEDULE_BY_BLOCK: List[Tuple[int, List[Competition]]] = [
        (
            0,
            [
                Competition(
                    COMPETITION_ID,
                    MODEL_CONSTRAINTS_BY_COMPETITION_ID[COMPETITION_ID],
                    1.0,
                )
            ],
        )
    ]

    def setUp(self):
        self.model_tracker = ModelTracker()
        self.local_store = DiskModelStore("test-models")
        self.remote_store = FakeRemoteModelStore()
        self.metadata_store = FakeModelMetadataStore()
        self.model_updater = ModelUpdater(
            metadata_store=self.metadata_store,
            remote_store=self.remote_store,
            local_store=self.local_store,
            model_tracker=self.model_tracker,
        )
        self.tiny_model = get_test_model()

    def tearDown(self):
        self.local_store.delete_unreferenced_models(dict(), 0)

    def test_get_metadata(self):
        hotkey = "test_hotkey"
        model_hash = "TestHash1"
        model_id = ModelId(
            namespace="TestPath",
            name="TestModel",
            competition_id=self.COMPETITION_ID,
            hash=model_hash,
            secure_hash=utils.get_hash_of_two_strings(model_hash, hotkey),
            commit="TestCommit",
        )
        model_metadata = ModelMetadata(id=model_id, block=1)

        asyncio.run(
            self.metadata_store.store_model_metadata_exact(hotkey, model_metadata)
        )

        metadata = asyncio.run(self.model_updater._get_metadata(hotkey))

        self.assertEqual(metadata.id, model_id)
        self.assertIsNotNone(metadata.block)

    def test_sync_model_bad_metadata(self):
        hotkey = "test_hotkey"
        model_hash = "TestHash1"
        model_id = ModelId(
            namespace="TestPath",
            name="TestModel",
            competition_id=self.COMPETITION_ID,
            hash=model_hash,
            secure_hash=utils.get_hash_of_two_strings(model_hash, hotkey),
            commit="TestCommit",
        )
        model_metadata = ModelMetadata(id=model_id, block=1)

        # Setup the metadata with a commit that doesn't exist in the remote store.
        asyncio.run(
            self.metadata_store.store_model_metadata_exact(hotkey, model_metadata)
        )

        # FakeRemoteModelStore raises a KeyError but HuggingFace may raise other exceptions.
        with self.assertRaises(Exception):
            asyncio.run(
                self.model_updater.sync_model(
                    hotkey=hotkey,
                    curr_block=100_000,
                    schedule_by_block=self.COMPETITION_SCHEDULE_BY_BLOCK,
                )
            )

    def test_sync_model_same_metadata(self):
        hotkey = "test_hotkey"
        model_hash = "TestHash1"
        model_id = ModelId(
            namespace="TestPath",
            name="TestModel",
            competition_id=self.COMPETITION_ID,
            hash=model_hash,
            secure_hash=utils.get_hash_of_two_strings(model_hash, hotkey),
            commit="TestCommit",
        )
        model_metadata = ModelMetadata(id=model_id, block=1)

        pt_model = self.tiny_model

        model = Model(id=model_id, pt_model=pt_model)

        # Setup the metadata, local, and model_tracker to match.
        asyncio.run(
            self.metadata_store.store_model_metadata_exact(hotkey, model_metadata)
        )
        self.local_store.store_model(hotkey, model)

        self.model_tracker.on_miner_model_updated(hotkey, model_metadata)

        asyncio.run(
            self.model_updater.sync_model(
                hotkey=hotkey,
                curr_block=100_000,
                schedule_by_block=self.COMPETITION_SCHEDULE_BY_BLOCK,
            )
        )

        # Tracker information did not change.
        self.assertEqual(
            self.model_tracker.get_model_metadata_for_miner_hotkey(hotkey),
            model_metadata,
        )

    def test_sync_model_new_metadata(self):
        hotkey = "test_hotkey"
        model_hash = "TestHash1"
        model_id = ModelId(
            namespace="TestPath",
            name="TestModel",
            competition_id=self.COMPETITION_ID,
            hash=model_hash,
            secure_hash=utils.get_hash_of_two_strings(model_hash, hotkey),
            commit="TestCommit",
        )
        model_metadata = ModelMetadata(id=model_id, block=1)

        pt_model = self.tiny_model

        model = Model(id=model_id, pt_model=pt_model)

        # Setup the metadata and remote store but not local or the model_tracker.
        asyncio.run(
            self.metadata_store.store_model_metadata_exact(hotkey, model_metadata)
        )
        asyncio.run(
            self.remote_store.upload_model(
                model,
                self.MODEL_CONSTRAINTS_BY_COMPETITION_ID.get(self.COMPETITION_ID),
            )
        )

        self.assertIsNone(
            self.model_tracker.get_model_metadata_for_miner_hotkey(hotkey)
        )

        # Our local store raises an exception from the Transformers.from_pretrained method if not found.
        with self.assertRaises(Exception):
            self.local_store.retrieve_model(hotkey, model_id, kwargs={})

        asyncio.run(
            self.model_updater.sync_model(
                hotkey=hotkey,
                curr_block=100_000,
                schedule_by_block=self.COMPETITION_SCHEDULE_BY_BLOCK,
            )
        )

        self.assertEqual(
            self.model_tracker.get_model_metadata_for_miner_hotkey(hotkey),
            model_metadata,
        )
        self.assertEqual(
            str(self.local_store.retrieve_model(hotkey, model_id, kwargs={})),
            str(model),
        )

    def test_sync_model_new_metadata_under_block_delay(self):
        hotkey = "test_hotkey"
        model_hash = "TestHash1"
        model_id = ModelId(
            namespace="TestPath",
            name="TestModel",
            competition_id=self.COMPETITION_ID,
            hash=model_hash,
            secure_hash=utils.get_hash_of_two_strings(model_hash, hotkey),
            commit="TestCommit",
        )
        model_metadata = ModelMetadata(id=model_id, block=1)

        pt_model = self.tiny_model

        model = Model(id=model_id, pt_model=pt_model)

        # Setup the metadata and remote store but not local or the model_tracker.
        asyncio.run(
            self.metadata_store.store_model_metadata_exact(hotkey, model_metadata)
        )
        asyncio.run(
            self.remote_store.upload_model(
                model,
                self.MODEL_CONSTRAINTS_BY_COMPETITION_ID.get(self.COMPETITION_ID),
            )
        )

        self.assertIsNone(
            self.model_tracker.get_model_metadata_for_miner_hotkey(hotkey)
        )

        updated = asyncio.run(
            self.model_updater.sync_model(
                hotkey=hotkey,
                curr_block=1,
                schedule_by_block=self.COMPETITION_SCHEDULE_BY_BLOCK,
            )
        )

        # Tracker information did not change.
        self.assertFalse(updated)
        self.assertIsNone(
            self.model_tracker.get_model_metadata_for_miner_hotkey(hotkey)
        )

    def test_sync_model_bad_hash(self):
        hotkey = "test_hotkey"
        model_hash = "TestHash1"
        model_id_chain = ModelId(
            namespace="TestPath",
            name="TestModel",
            competition_id=self.COMPETITION_ID,
            hash=model_hash,
            secure_hash=utils.get_hash_of_two_strings(model_hash, hotkey),
            commit="TestCommit",
        )

        model_metadata = ModelMetadata(id=model_id_chain, block=1)
        bad_hash = "BadHash"
        model_id = ModelId(
            namespace="TestPath",
            name="TestModel",
            competition_id=self.COMPETITION_ID,
            hash=bad_hash,
            secure_hash=utils.get_hash_of_two_strings(bad_hash, hotkey),
            commit="TestCommit",
        )

        pt_model = self.tiny_model

        model = Model(id=model_id, pt_model=pt_model)

        # Setup the metadata and remote store and but not local or the model tracker.
        asyncio.run(
            self.metadata_store.store_model_metadata_exact(hotkey, model_metadata)
        )
        self.remote_store.inject_mismatched_model(model_id_chain, model)

        # Assert we fail due to the hash mismatch between the model in remote store and the metadata on chain.
        with self.assertRaises(MinerMisconfiguredError) as context:
            asyncio.run(
                self.model_updater.sync_model(
                    hotkey=hotkey,
                    curr_block=100_000,
                    schedule_by_block=self.COMPETITION_SCHEDULE_BY_BLOCK,
                )
            )

        self.assertIn("Hash", str(context.exception))

        # Also assert that the model tracker is tracking the model even after the exception above.
        self.assertEqual(
            self.model_tracker.get_model_metadata_for_miner_hotkey(hotkey),
            model_metadata,
        )

    # TODO: Create separate tests for size, architecture, and eps.
    def test_sync_model_wrong_parameters(self):
        hotkey = "test_hotkey"
        model_hash = "TestHash1"
        model_id = ModelId(
            namespace="TestPath",
            name="TestModel",
            competition_id=self.COMPETITION_ID,
            hash=model_hash,
            secure_hash=utils.get_hash_of_two_strings(model_hash, hotkey),
            commit="TestCommit",
        )
        model_metadata = ModelMetadata(id=model_id, block=1)

        config = GPT2Config(
            n_head=10,
            n_layer=12,
            n_embd=760,
        )
        pt_model = GPT2LMHeadModel(config)

        model = Model(id=model_id, pt_model=pt_model)

        # Setup the metadata and remote store but not local or the model_tracker.
        asyncio.run(
            self.metadata_store.store_model_metadata_exact(hotkey, model_metadata)
        )
        asyncio.run(
            self.remote_store.upload_model(
                model,
                self.MODEL_CONSTRAINTS_BY_COMPETITION_ID.get(self.COMPETITION_ID),
            ),
        )

        # Assert we fail due to not meeting the competition parameters.
        with self.assertRaises(MinerMisconfiguredError) as context:
            asyncio.run(
                self.model_updater.sync_model(
                    hotkey=hotkey,
                    curr_block=100_000,
                    schedule_by_block=self.COMPETITION_SCHEDULE_BY_BLOCK,
                )
            )

        self.assertIn("does not satisfy parameters", str(context.exception))

        # Also assert that the model tracker is tracking the model even after the exception above.
        self.assertEqual(
            self.model_tracker.get_model_metadata_for_miner_hotkey(hotkey),
            model_metadata,
        )

    # TODO: Create test for valid competition at too early of a block once added.


if __name__ == "__main__":
    unittest.main()



import copy
import pickle
import threading
from collections import defaultdict
from typing import Dict, List, Optional, Set

import bittensor as bt

from taoverse.model.data import EvalResult, ModelMetadata


class ModelTracker:
    """Tracks model information for each miner.

    Thread safe.
    """

    _MAX_EVAL_HISTORY_LEN = 3

    def __init__(
        self,
        max_eval_history_len: int = _MAX_EVAL_HISTORY_LEN,
    ):
        self.max_eval_history_len = max_eval_history_len

        # Create a dict from miner hotkey to model metadata.
        self.miner_hotkey_to_model_metadata_dict = dict()
        # Sadly, a defaultdict of defaultdict is not picklable, so we have to use a regular dict.
        # This dictionary is a map from hotkey to competition id to a list of eval results.
        self.miner_hotkey_to_eval_results: Dict[Dict[str, List[EvalResult]]] = defaultdict(dict)

        # Make this class thread safe because it will be accessed by multiple threads.
        # One for the downloading new models loop and one for the validating models loop.
        self.lock = threading.RLock()

    def save_state(self, filepath):
        """Save the current state to the provided filepath."""

        # Open a writable binary file for pickle.
        with self.lock:
            with open(filepath, "wb") as f:
                pickle.dump(
                    [
                        self.miner_hotkey_to_model_metadata_dict,
                        self.miner_hotkey_to_eval_results,
                    ],
                    f,
                )

    def load_state(self, filepath):
        """Load the state from the provided filepath."""

        # Open a readable binary file for pickle.
        with open(filepath, "rb") as f:
            (
                self.miner_hotkey_to_model_metadata_dict,
                self.miner_hotkey_to_eval_results,
            ) = pickle.load(f)

    def get_miner_hotkey_to_model_metadata_dict(self) -> Dict[str, ModelMetadata]:
        """Returns the mapping from miner hotkey to model metadata."""

        # Return a copy to ensure outside code can't modify the scores.
        with self.lock:
            return copy.deepcopy(self.miner_hotkey_to_model_metadata_dict)

    def get_model_metadata_for_miner_hotkey(
        self, hotkey: str
    ) -> Optional[ModelMetadata]:
        """Returns the model metadata for a given hotkey if any."""

        with self.lock:
            if hotkey in self.miner_hotkey_to_model_metadata_dict:
                return self.miner_hotkey_to_model_metadata_dict[hotkey]
            return None

    def get_eval_results_for_miner_hotkey(self, hotkey: str, competition_id: int) -> List[EvalResult]:
        """Returns the latest evaluation results for a miner, ordered by block of eval (oldest first)."""

        with self.lock:
            return copy.deepcopy(self.miner_hotkey_to_eval_results[hotkey].get(competition_id, []))
        
    def clear_eval_results(self, competition_id: int) -> None:
        """Clears all evaluation results for a given competition id."""
        
        with self.lock:
            for eval_results in self.miner_hotkey_to_eval_results.values():
                if competition_id in eval_results:
                    del eval_results[competition_id]
        
    def get_block_last_evaluated(self, hotkey: str) -> Optional[int]:
        """Returns the block of the most recent evaluation for a miner, across all competitions.
        
        Args:
            hotkey (str): The miner's hotkey.
            
        Returns:
            Optional[int]: The block of the most recent evaluation for the miner, or None if no evaluations have been made.
        """

        with self.lock:
            eval_blocks = [
                eval_result.block for eval_results in self.miner_hotkey_to_eval_results[hotkey].values()
                for eval_result in eval_results
            ]
            return sorted(eval_blocks)[-1] if eval_blocks else None

    def on_hotkeys_updated(self, incoming_hotkeys: Set[str]):
        """Notifies the tracker which hotkeys are currently being tracked on the metagraph."""

        with self.lock:
            existing_hotkeys = set(self.miner_hotkey_to_model_metadata_dict.keys())
            for hotkey in existing_hotkeys - incoming_hotkeys:
                del self.miner_hotkey_to_model_metadata_dict[hotkey]
                del self.miner_hotkey_to_eval_results[hotkey]
                bt.logging.trace(f"Removed outdated hotkey: {hotkey} from ModelTracker")

    def on_miner_model_updated(
        self,
        hotkey: str,
        model_metadata: ModelMetadata,
    ) -> None:
        """Notifies the tracker that a miner has had their associated model updated.

        Args:
            hotkey (str): The miner's hotkey.
            model_metadata (ModelMetadata): The latest model metadata of the miner.
        """
        with self.lock:
            prev_metadata = self.miner_hotkey_to_model_metadata_dict.get(hotkey, None)
            self.miner_hotkey_to_model_metadata_dict[hotkey] = model_metadata

            # If the model was updated, clear the evaluation results since they're no
            # longer relevant.
            if prev_metadata != model_metadata:
                self.miner_hotkey_to_eval_results[hotkey].clear()

            bt.logging.trace(f"Updated Miner {hotkey}. ModelMetadata={model_metadata}.")

    def on_model_evaluated(self, hotkey: str, competition_id: int, result: EvalResult) -> None:
        """Notifies the tracker that a model has been evaluated.

        Args:
            hotkey (str): The miner's hotkey.
            competition_id (int): The competition id for this eval result.
            result (EvalResult): The evaluation result.
        """
        with self.lock:
            eval_results = self.miner_hotkey_to_eval_results[hotkey].get(competition_id, [])
            eval_results.append(result)
            eval_results.sort(key=lambda x: x.block)

            if len(eval_results) > self.max_eval_history_len:
                eval_results.pop(0)
            self.miner_hotkey_to_eval_results[hotkey][competition_id] = eval_results

            bt.logging.trace(f"Updated eval results on {hotkey} for comp {competition_id}. EvalResult={result}.")



import abc
from typing import Any, Dict

from taoverse.model.data import Model, ModelId


class LocalModelStore(abc.ABC):
    """An abstract base class for storing and retrieving a pre trained model locally."""

    @abc.abstractmethod
    def store_model(self, hotkey: str, model: Model) -> ModelId:
        """Stores a trained model in the appropriate location based on implementation."""
        pass

    @abc.abstractmethod
    def get_path(self, hotkey: str) -> str:
        """Returns the path to the appropriate location based on implementation."""
        pass

    @abc.abstractmethod
    def retrieve_model(
        self, hotkey: str, model_id: ModelId, kwargs: Dict[str, Any]
    ) -> Model:
        """Retrieves a trained model from the appropriate location based on implementation."""
        pass

    @abc.abstractmethod
    def delete_unreferenced_models(
        self, valid_models_by_hotkey: Dict[str, ModelId], grace_period_seconds: int
    ):
        """Check across all of local storage and delete unreferenced models out of grace period."""
        pass



import os
import unittest

import torch
from transformers import LlamaForCausalLM

from taoverse.model.competition.competition_tracker import CompetitionTracker
from taoverse.model.competition.data import Competition, ModelConstraints
from taoverse.model.competition.epsilon import FixedEpsilon


class TestCompetitionTracker(unittest.TestCase):
    COMPETITION_1_PARAMETERS = Competition(
        id=1,
        constraints=ModelConstraints(
            max_model_parameter_size=8 * 1024 * 1024 * 1024,
            sequence_length=4096,
            allowed_architectures=[LlamaForCausalLM],
            tokenizer="Xenova/gpt-4",
            kwargs={
                "torch_dtype": torch.bfloat16,
            },
            eval_block_delay=1200,  # ~4 hours.
            epsilon_func=FixedEpsilon(0.005),
            max_bytes=15 * 1024 * 1024 * 1024,  # 15 GB
        ),
        reward_percentage=0.6,
    )
    COMPETITION_2_PARAMETERS = Competition(
        id=2,
        constraints=ModelConstraints(
            max_model_parameter_size=2 * 1024 * 1024 * 1024,
            sequence_length=2048,
            allowed_architectures=[LlamaForCausalLM],
            tokenizer="Xenova/gpt-4",
            kwargs={
                "torch_dtype": torch.bfloat16,
            },
            eval_block_delay=1200,  # ~4 hours.
            epsilon_func=FixedEpsilon(0.005),
            max_bytes=15 * 1024 * 1024 * 1024,  # 15 GB
        ),
        reward_percentage=0.4,
    )

    def setUp(self):
        self.num_neurons = 4
        self.alpha = 0.5
        self.competition_tracker = CompetitionTracker(self.num_neurons, self.alpha)

    def test_record_competition_weights_new(self):
        self.competition_tracker.record_competition_weights(
            1, torch.Tensor([0.1, 0.2, 0.3, 0.4])
        )

        # Since this is a net new competition, check that weights go immediately to the new values.
        self.assertTrue(
            torch.equal(
                self.competition_tracker.weights_by_competition[1],
                torch.Tensor([0.1, 0.2, 0.3, 0.4]),
            )
        )

    def test_record_competition_weights_normalized(self):
        self.competition_tracker.record_competition_weights(
            1, torch.Tensor([0.2, 0.4, 0.6, 0.8])
        )

        # Check that the weights are normalized to sum to 1.
        self.assertTrue(
            torch.equal(
                self.competition_tracker.weights_by_competition[1],
                torch.Tensor([0.1, 0.2, 0.3, 0.4]),
            )
        )

    def test_record_competition_weights_moving_average(self):
        self.competition_tracker.record_competition_weights(
            1, torch.Tensor([1, 0, 0, 0])
        )
        self.competition_tracker.record_competition_weights(
            1, torch.Tensor([0, 0, 0, 1])
        )

        # Check that the weights are a moving average according to the alpha of 0.5.
        self.assertTrue(
            torch.equal(
                self.competition_tracker.weights_by_competition[1],
                torch.Tensor([0.5, 0, 0, 0.5]),
            )
        )

    def test_get_competition_weights(self):
        self.competition_tracker.record_competition_weights(
            1, torch.Tensor([0.1, 0.2, 0.3, 0.4])
        )

        weights = self.competition_tracker.get_competition_weights(1)

        self.assertTrue(torch.equal(weights, torch.Tensor([0.1, 0.2, 0.3, 0.4])))

    def test_get_subnet_weights_one_competition(self):
        self.competition_tracker.record_competition_weights(
            1, torch.Tensor([0.1, 0.2, 0.3, 0.4])
        )

        weights = self.competition_tracker.get_subnet_weights(
            [self.COMPETITION_1_PARAMETERS]
        )

        self.assertTrue(torch.equal(weights, torch.Tensor([0.1, 0.2, 0.3, 0.4])))

    def test_get_subnet_weights_one_competition_with_threshold(self):
        self.competition_tracker.record_competition_weights(
            1, torch.Tensor([0.1, 0.1, 0.2, 0.3, 0.3])
        )

        # Only count uids with at least 0.3 weight for this competition.
        weights = self.competition_tracker.get_subnet_weights(
            [self.COMPETITION_1_PARAMETERS], min_comp_weight_threshold=0.3
        )

        # On normalization expect the two remaining 0.3s to split the weight evenly.
        self.assertTrue(torch.equal(weights, torch.Tensor([0.0, 0.0, 0.0, 0.5, 0.5])))

    def test_get_subnet_weights_two_competitions(self):
        self.competition_tracker.record_competition_weights(
            1, torch.Tensor([1, 1, 0, 0])
        )
        self.competition_tracker.record_competition_weights(
            2, torch.Tensor([0, 0, 5, 5])
        )

        weights = self.competition_tracker.get_subnet_weights(
            [self.COMPETITION_1_PARAMETERS, self.COMPETITION_2_PARAMETERS]
        )

        # Check that the weights are both normalized and rewarded according to competition reward percent.
        self.assertTrue(torch.equal(weights, torch.Tensor([0.3, 0.3, 0.2, 0.2])))

    def test_get_subnet_weights_two_competitions_with_threshold(self):
        self.competition_tracker.record_competition_weights(
            1, torch.Tensor([1, 2, 0, 0])
        )
        self.competition_tracker.record_competition_weights(
            2, torch.Tensor([0, 0, 5, 5])
        )

        # After normalization only the 2nd uid from comp 1 should have any weight.
        weights = self.competition_tracker.get_subnet_weights(
            [self.COMPETITION_1_PARAMETERS, self.COMPETITION_2_PARAMETERS],
            min_comp_weight_threshold=0.4,
        )

        # Check that the weights are both normalized and rewarded according to competition reward percent.
        # Although only comp 1 was trimmed the remaining uid should still receive the full 60%.
        self.assertTrue(torch.equal(weights, torch.Tensor([0.0, 0.6, 0.2, 0.2])))

    def test_resize_one_competition(self):
        self.competition_tracker.record_competition_weights(
            1, torch.Tensor([0.1, 0.2, 0.3, 0.2, 0.2])
        )

        # Check that the internal state immediately expands to 5 neurons.
        self.assertTrue(
            torch.equal(
                self.competition_tracker.weights_by_competition[1],
                torch.Tensor(
                    [0.1, 0.2, 0.3, 0.2, 0.2],
                ),
            )
        )

    def test_resize_two_competition(self):
        self.competition_tracker.record_competition_weights(
            1, torch.Tensor([0.1, 0.2, 0.3, 0.4])
        )
        self.competition_tracker.record_competition_weights(
            2, torch.Tensor([0.1, 0.2, 0.3, 0.2, 0.2])
        )

        # Check that the internal state of the first competition is expanded with 0s.
        self.assertTrue(
            torch.equal(
                self.competition_tracker.weights_by_competition[1],
                torch.Tensor([0.1, 0.2, 0.3, 0.4, 0.0]),
            )
        )
        self.assertTrue(
            torch.equal(
                self.competition_tracker.weights_by_competition[2],
                torch.Tensor([0.1, 0.2, 0.3, 0.2, 0.2]),
            )
        )

    def test_reset_competitions(self):
        self.competition_tracker.record_competition_weights(
            1, torch.Tensor([1, 0, 0, 0])
        )
        self.competition_tracker.record_competition_weights(
            2, torch.Tensor([0, 0, 0, 1])
        )

        self.competition_tracker.reset_competitions({1})

        # Check that the weights for competition 2 are no longer tracked.
        self.assertTrue(1 in self.competition_tracker.weights_by_competition)
        self.assertFalse(2 in self.competition_tracker.weights_by_competition)

    def test_roundtrip_state(self):
        self.competition_tracker.record_competition_weights(
            1, torch.Tensor([1, 0, 0, 0])
        )

        state_path = ".test_competition_tracker_state.pickle"
        self.competition_tracker.save_state(state_path)

        new_competition_tracker = CompetitionTracker(num_neurons=0)
        new_competition_tracker.load_state(state_path)

        os.remove(state_path)

        self.assertEqual(
            self.competition_tracker.num_neurons,
            new_competition_tracker.num_neurons,
        )
        self.assertEqual(
            self.competition_tracker.alpha,
            new_competition_tracker.alpha,
        )
        self.assertEqual(
            len(new_competition_tracker.weights_by_competition),
            1,
        )
        self.assertTrue(
            torch.equal(
                self.competition_tracker.weights_by_competition[1],
                new_competition_tracker.weights_by_competition[1],
            )
        )


if __name__ == "__main__":
    unittest.main()



import asyncio
import functools
import os
from typing import Optional

import bittensor as bt

from taoverse.model.data import ModelId, ModelMetadata
from taoverse.model.storage.model_metadata_store import ModelMetadataStore
from taoverse.utilities import utils


class ChainModelMetadataStore(ModelMetadataStore):
    """Chain based implementation for storing and retrieving metadata about a model."""

    def __init__(
        self,
        subtensor: bt.subtensor,
        subnet_uid: int,
        # Wallet is only needed to write to the chain, not to read.
        wallet: Optional[bt.wallet] = None,
    ):
        self.subtensor = subtensor
        self.wallet = wallet
        self.subnet_uid = subnet_uid

    async def store_model_metadata(
        self,
        hotkey: str,
        model_id: ModelId,
        wait_for_inclusion: bool = True,
        wait_for_finalization: bool = True,
    ):
        """Stores model metadata on this subnet for a specific wallet."""
        if self.wallet is None:
            raise ValueError("No wallet available to write to the chain.")

        data = model_id.to_compressed_str()

        # Wrap calls to the subtensor in a thread with a timeout to handle potential hangs.
        partial = functools.partial(
            bt.core.extrinsics.serving.publish_metadata,
            self.subtensor,
            self.wallet,
            self.subnet_uid,
            f"Raw{len(data)}",
            data.encode(),
            wait_for_inclusion,
            wait_for_finalization,
        )
        utils.run_in_thread(partial, 30)

    async def retrieve_model_metadata(self, hotkey: str) -> Optional[ModelMetadata]:
        """Retrieves model metadata on this subnet for specific hotkey"""

        # Wrap calls to the subtensor in a thread with a timeout to handle potential hangs.
        partial = functools.partial(
            bt.core.extrinsics.serving.get_metadata,
            self.subtensor,
            self.subnet_uid,
            hotkey,
        )
        metadata = utils.run_in_thread(partial, 30)

        if not metadata:
            return None

        commitment = metadata["info"]["fields"][0]
        hex_data = commitment[list(commitment.keys())[0]][2:]

        chain_str = bytes.fromhex(hex_data).decode()

        model_id = None

        try:
            model_id = ModelId.from_compressed_str(chain_str)
        except:
            # If the metadata format is not correct on the chain then we return None.
            bt.logging.trace(
                f"Failed to parse the metadata on the chain for hotkey {hotkey}."
            )
            return None

        model_metadata = ModelMetadata(id=model_id, block=metadata["block"])

        return model_metadata


# Can only commit data every ~20 minutes.
async def test_store_model_metadata():
    """Verifies that the ChainModelMetadataStore can store data on the chain."""
    model_id = ModelId(
        namespace="TestPath",
        name="TestModel",
        competition_id=1,
        hash="TestHash1",
        commit="1.0",
    )

    # Use a different subnet that does not leverage chain storage to avoid conflicts.
    # TODO switch to a mocked version when it supports commits.
    subtensor = bt.subtensor()

    # Uses .env configured wallet/hotkey/uid for the test.
    coldkey = os.getenv("TEST_COLDKEY")
    hotkey = os.getenv("TEST_HOTKEY")
    net_uid = int(os.getenv("TEST_SUBNET_UID"))

    wallet = bt.wallet(name=coldkey, hotkey=hotkey)

    metadata_store = ChainModelMetadataStore(
        subtensor=subtensor,
        subnet_uid=net_uid,
        wallet=wallet,
    )

    # Store the metadata on chain.
    await metadata_store.store_model_metadata(
        hotkey=hotkey,
        model_id=model_id,
        wait_for_inclusion=True,
        wait_for_finalization=False,
    )

    print(f"Finished storing {model_id} on the chain.")


async def test_retrieve_model_metadata():
    """Verifies that the ChainModelMetadataStore can retrieve data from the chain."""
    expected_model_id = ModelId(
        namespace="TestPath",
        name="TestModel",
        competition_id=1,
        hash="TestHash1",
        commit="1.0",
    )

    # Use a different subnet that does not leverage chain storage to avoid conflicts.
    # TODO switch to a mocked version when it supports commits.
    subtensor = bt.subtensor()

    # Uses .env configured hotkey/uid for the test.
    net_uid = int(os.getenv("TEST_SUBNET_UID"))
    hotkey_address = os.getenv("TEST_HOTKEY_ADDRESS")

    # Do not require a wallet for retrieving data.
    metadata_store = ChainModelMetadataStore(
        subtensor=subtensor,
        subnet_uid=net_uid,
        wallet=None,
    )

    # Retrieve the metadata from the chain.
    model_metadata = await metadata_store.retrieve_model_metadata(hotkey_address)

    print(f"Expecting matching model id: {expected_model_id == model_metadata.id}")


# Can only commit data every ~20 minutes.
async def test_roundtrip_model_metadata():
    """Verifies that the ChainModelMetadataStore can roundtrip data on the chain."""
    model_id = ModelId(
        namespace="TestPath",
        name="TestModel",
        competition_id=1,
        hash="TestHash1",
        commit="1.0",
    )

    # Use a different subnet that does not leverage chain storage to avoid conflicts.
    # TODO switch to a mocked version when it supports commits.
    subtensor = bt.subtensor()

    # Uses .env configured wallet/hotkey/uid for the test.
    coldkey = os.getenv("TEST_COLDKEY")
    hotkey = os.getenv("TEST_HOTKEY")
    net_uid = int(os.getenv("TEST_SUBNET_UID"))

    wallet = bt.wallet(name=coldkey, hotkey=hotkey)

    metadata_store = ChainModelMetadataStore(
        subtensor=subtensor,
        subnet_uid=net_uid,
        wallet=wallet,
    )

    # Store the metadata on chain.
    await metadata_store.store_model_metadata(
        hotkey=hotkey,
        model_id=model_id,
        wait_for_inclusion=True,
        wait_for_finalization=True,
    )

    # Retrieve the metadata from the chain.
    model_metadata = await metadata_store.retrieve_model_metadata(hotkey)

    print(f"Expecting matching metadata: {model_id == model_metadata.id}")


if __name__ == "__main__":
    # Can only commit data every ~20 minutes.
    asyncio.run(test_roundtrip_model_metadata())
    asyncio.run(test_store_model_metadata())
    asyncio.run(test_retrieve_model_metadata())



from typing import List, Set, Tuple

import bittensor as bt


def assert_registered(wallet: bt.wallet, metagraph: bt.metagraph) -> int:
    """Asserts the wallet is a registered miner and returns the miner's UID.

    Raises:
        ValueError: If the wallet is not registered.
    """
    if wallet.hotkey.ss58_address not in metagraph.hotkeys:
        raise ValueError(
            f"You are not registered. \nUse: \n`btcli s register --netuid {metagraph.netuid}` to register via burn \n or btcli s pow_register --netuid {metagraph.netuid} to register with a proof of work"
        )
    uid = metagraph.hotkeys.index(wallet.hotkey.ss58_address)
    bt.logging.success(
        f"You are registered with address: {wallet.hotkey.ss58_address} and uid: {uid}"
    )

    return uid


def get_top_miners(
    metagraph: bt.metagraph, min_vali_stake: int, min_miner_weight_percent: float
) -> Set[int]:
    """Returns the set of top miners, chosen based on weights set on the valis above the specifed threshold.

    Args:
        metagraph (bt.metagraph): Metagraph to use. Must not be lite.
        min_vali_stake (int): Minimum stake threshold for a vali's weights to be considered.
        min_miner_weight_percent (float): Minimum weight on a vali for the miner to count as a top miner.
    """

    top_miners = set()

    # Find validators over 100k in stake.
    valis_by_stake = get_high_stake_validators(metagraph, min_vali_stake)

    # For each, find miners with at least min_miner_weight_percent of the weights.
    # Since there can be multiple competitions at different reward percentages we can't just check biggest.
    for uid in valis_by_stake:
        # Weights is a list of (uid, weight) pairs
        weights: List[Tuple[int, float]] = metagraph.neurons[uid].weights
        total_weight = sum(weight for _, weight in weights)

        threshold = total_weight * min_miner_weight_percent
        for uid, weight in weights:
            if weight >= threshold:
                top_miners.add(uid)

    return top_miners


def get_high_stake_validators(metagraph: bt.metagraph, min_stake: int) -> Set[int]:
    """Returns a set of validators at or above the specified stake threshold for the subnet"""
    valis = set()

    for uid, stake in enumerate(metagraph.S):
        # Use vPermit to check for validators rather than vTrust because we'd rather
        # cast a wide net in the case that vTrust is 0 due to an unhealthy state of the
        # subnet.
        if stake >= min_stake and metagraph.validator_permit[uid]:
            valis.add(uid)

    return valis


def get_hash_of_sync_block(subtensor: bt.subtensor, sync_cadence: int) -> int:
    """Returns the hash of the most recent block that is a multiple of the sync cadence."""
    current_block = subtensor.get_current_block()
    sync_block = current_block // sync_cadence * sync_cadence
    return hash(subtensor.get_block_hash(sync_block))



from dataclasses import dataclass, field
import dataclasses
from typing import Any, List, Optional, Type

from transformers import PreTrainedModel

from taoverse.model.competition.epsilon import EpsilonFunc
from taoverse.model.eval.task import EvalTask


@dataclass
class NormValidationConstraints:
    """Defines the validation constriants to be checked based on model norms."""

    norm_eps_soft: int
    norm_eps_soft_percent_threshold: float
    norm_eps_hard: int


@dataclass
class ModelConstraints:
    """Defines the constraints for models submitted to a specific competition."""

    # The maximum parameter size allowed for models
    max_model_parameter_size: int

    # Architecture class of model
    allowed_architectures: List[Type[PreTrainedModel]]

    # The model's sequence length.
    sequence_length: int

    # Block delay before evaluating uploaded models. Based on look-back period for eval data collection.
    eval_block_delay: int

    # The function to compute epsilon as a function of the model's submitted block and the current block.
    epsilon_func: EpsilonFunc = field(compare=False)

    # The maximum size (in bytes) the model is allowed to be.
    max_bytes: int

    # Any additional arguments to pass to from_pretrained
    kwargs: Any = field(default_factory=dict)

    # The Pretrained tokenizer to use. If not set then then any tokenizer is allowed.
    tokenizer: Optional[str] = None

    # The minimum parameter size allowed for models
    min_model_parameter_size: int = 0

    # Norm validation values.
    norm_validation_constraints: Optional[NormValidationConstraints] = None


@dataclass
class Competition:
    """Defines a competition."""

    # Unique ID for this competition.
    # Recommend making an IntEnum for use in the subnet codebase.
    id: int

    # All restrictions on models allowed in this competition.
    constraints: ModelConstraints

    # Percentage of emissions dedicated to this competition.
    reward_percentage: float

    # The set of tasks used to evaluate models in this competition.
    eval_tasks: List[EvalTask] = dataclasses.field(default_factory=list)



import unittest

from transformers import AutoTokenizer

import taoverse.model.storage.disk.utils as utils
from taoverse.model.data import Model, ModelId
from taoverse.model.storage.disk.disk_model_store import DiskModelStore
from tests.utils import get_test_model


class TestDiskModelStore(unittest.TestCase):
    def setUp(self):
        self.base_dir = "test-models"
        self.disk_store = DiskModelStore(self.base_dir)
        self.tiny_model = get_test_model()

    def tearDown(self):
        # Clean up the model directory.
        self.disk_store.delete_unreferenced_models(dict(), 0)

    def test_get_path(self):
        hotkey = "hotkey0"

        expected_path = utils.get_local_miner_dir("test-models", hotkey)
        actual_path = self.disk_store.get_path(hotkey)

        self.assertEqual(expected_path, actual_path)

    def test_store_and_retrieve_model(self):
        hotkey = "hotkey0"
        model_id = ModelId(
            namespace="TestPath",
            name="TestModel",
            competition_id=1,
            secure_hash="TestHash1",
            commit="TestCommit",
        )

        pt_model = self.tiny_model

        model = Model(id=model_id, pt_model=pt_model)

        # Store the model locally.
        self.disk_store.store_model(hotkey, model)

        # Retrieve the model locally.
        retrieved_model = self.disk_store.retrieve_model(hotkey, model_id, kwargs={})

        # Check that they match.
        self.assertEqual(str(model), str(retrieved_model))

    def test_store_and_retrieve_model_with_tokenizer(self):
        hotkey = "hotkeyTok"
        model_id = ModelId(
            namespace="TestPathTok",
            name="TestModelTok",
            competition_id=1,
            secure_hash="TestHash1Tok",
            commit="TestCommitTok",
        )

        pt_model = self.tiny_model
        tokenizer = AutoTokenizer.from_pretrained("Xenova/gpt-4")

        model = Model(id=model_id, pt_model=pt_model, tokenizer=tokenizer)

        # Store the model locally.
        self.disk_store.store_model(hotkey, model)

        # Retrieve the model locally.
        retrieved_model = self.disk_store.retrieve_model(hotkey, model_id, kwargs={})

        # Overwrite the name of the tokenizer to avoid it using the local path.
        retrieved_model.tokenizer.name_or_path = "Xenova/gpt-4"

        # Check that they match.
        self.assertEqual(str(model), str(retrieved_model))

    def test_delete_unreferenced_models(self):
        hotkey = "hotkey0"

        # Make 2 model ids with different hashes / commits.
        model_id_1 = ModelId(
            namespace="TestPath",
            name="TestModel",
            competition_id=1,
            secure_hash="TestHash1",
            commit="TestCommit",
        )
        model_id_2 = ModelId(
            namespace="TestPath",
            name="TestModel",
            competition_id=1,
            secure_hash="TestHash2",
            commit="TestCommit2",
        )

        pt_model = self.tiny_model

        model_1 = Model(id=model_id_1, pt_model=pt_model)
        model_2 = Model(id=model_id_2, pt_model=pt_model)

        # Store both models locally.
        self.disk_store.store_model(hotkey, model_1)
        self.disk_store.store_model(hotkey, model_2)

        # Create the mapping of hotkey to model_id with only the 2nd model.
        valid_models_by_hotkey = dict()
        valid_models_by_hotkey[hotkey] = model_id_2

        # Clear the unreferenced models
        self.disk_store.delete_unreferenced_models(valid_models_by_hotkey, 0)

        # Confirm that model 1 is deleted
        with self.assertRaises(Exception):
            self.disk_store.retrieve_model(hotkey, model_id_1, kwargs={})

        # Confirm that model 2 is still there
        model_2_retrieved = self.disk_store.retrieve_model(
            hotkey, model_id_2, kwargs={}
        )
        self.assertEqual(str(model_2), str(model_2_retrieved))

    def test_delete_unreferenced_models_removed_hotkey(self):
        hotkey_1 = "hotkey1"
        hotkey_2 = "hotkey2"

        # Make 2 model ids with different hashes / commits.
        model_id_1 = ModelId(
            namespace="TestPath",
            name="TestModel",
            competition_id=1,
            secure_hash="TestHash1",
            commit="TestCommit",
        )
        model_id_2 = ModelId(
            namespace="TestPath",
            name="TestModel",
            competition_id=1,
            secure_hash="TestHash2",
            commit="TestCommit2",
        )

        pt_model = self.tiny_model

        model_1 = Model(id=model_id_1, pt_model=pt_model)
        model_2 = Model(id=model_id_2, pt_model=pt_model)

        # Store both models locally.
        self.disk_store.store_model(hotkey_1, model_1)
        self.disk_store.store_model(hotkey_2, model_2)

        # Create the mapping of hotkey to model_id with only the 2nd model.
        valid_models_by_hotkey = dict()
        valid_models_by_hotkey[hotkey_2] = model_id_2

        # Clear the unreferenced models
        self.disk_store.delete_unreferenced_models(valid_models_by_hotkey, 0)

        # Confirm that model 1 is deleted
        with self.assertRaises(Exception):
            self.disk_store.retrieve_model(hotkey_1, model_id_1, kwargs={})

        # Confirm that model 2 is still there
        model_2_retrieved = self.disk_store.retrieve_model(
            hotkey_2, model_id_2, kwargs={}
        )
        self.assertEqual(str(model_2), str(model_2_retrieved))

    def test_delete_unreferenced_models_in_grace(self):
        hotkey = "hotkey0"

        # Make 2 model ids with different hashes / commits.
        model_id_1 = ModelId(
            namespace="TestPath",
            name="TestModel",
            competition_id=1,
            secure_hash="TestHash1",
            commit="TestCommit",
        )
        model_id_2 = ModelId(
            namespace="TestPath",
            name="TestModel",
            competition_id=1,
            secure_hash="TestHash2",
            commit="TestCommit2",
        )

        pt_model = self.tiny_model

        model_1 = Model(id=model_id_1, pt_model=pt_model)
        model_2 = Model(id=model_id_2, pt_model=pt_model)

        # Store both models locally.
        self.disk_store.store_model(hotkey, model_1)
        self.disk_store.store_model(hotkey, model_2)

        # Create the mapping of hotkey to model_id with only the 2nd model.
        valid_models_by_hotkey = dict()
        valid_models_by_hotkey[hotkey] = model_id_2

        # Clear the unreferenced models
        self.disk_store.delete_unreferenced_models(valid_models_by_hotkey, 60)

        # Confirm that model 1 is still there.
        model_1_retrieved = self.disk_store.retrieve_model(
            hotkey, model_id_1, kwargs={}
        )
        self.assertEqual(str(model_1), str(model_1_retrieved))

        # Confirm that model 2 is still there
        model_2_retrieved = self.disk_store.retrieve_model(
            hotkey, model_id_2, kwargs={}
        )
        self.assertEqual(str(model_2), str(model_2_retrieved))

    def test_delete_unreferenced_models_removed_hotkey_in_grace(self):
        hotkey_1 = "hotkey1"
        hotkey_2 = "hotkey2"

        # Make 2 model ids with different hashes / commits.
        model_id_1 = ModelId(
            namespace="TestPath",
            name="TestModel",
            competition_id=1,
            secure_hash="TestHash1",
            commit="TestCommit",
        )
        model_id_2 = ModelId(
            namespace="TestPath",
            name="TestModel",
            competition_id=1,
            secure_hash="TestHash2",
            commit="TestCommit2",
        )

        pt_model = self.tiny_model

        model_1 = Model(id=model_id_1, pt_model=pt_model)
        model_2 = Model(id=model_id_2, pt_model=pt_model)

        # Store both models locally.
        self.disk_store.store_model(hotkey_1, model_1)
        self.disk_store.store_model(hotkey_2, model_2)

        # Create the mapping of hotkey to model_id with only the 2nd model.
        valid_models_by_hotkey = dict()
        valid_models_by_hotkey[hotkey_2] = model_id_2

        # Clear the unreferenced models
        self.disk_store.delete_unreferenced_models(valid_models_by_hotkey, 60)

        # Confirm that model 1 is still there
        model_1_retrieved = self.disk_store.retrieve_model(
            hotkey_1, model_id_1, kwargs={}
        )
        self.assertEqual(str(model_1), str(model_1_retrieved))

        # Confirm that model 2 is still there
        model_2_retrieved = self.disk_store.retrieve_model(
            hotkey_2, model_id_2, kwargs={}
        )
        self.assertEqual(str(model_2), str(model_2_retrieved))


if __name__ == "__main__":
    unittest.main()



import typing

import bittensor as bt
import torch

from taoverse.model.competition.data import Competition


class CompetitionTracker:
    """Tracks weights of each miner (by UID) per competition."""

    def __init__(
        self,
        num_neurons: int,
        alpha: float = 0.5,
    ):
        self.weights_by_competition: typing.Dict[int, torch.Tensor] = {}
        self.num_neurons = num_neurons
        self.alpha = alpha

    # TODO: Consider a record_competition_results which then does additional processing into weights.

    def record_competition_weights(
        self, competition_id: int, new_weights: torch.Tensor
    ):
        """Records the weights from a new run of a competition, updating a moving average.

        Args:
            competition_id (int): Which competition the weights are for.
            new_weights (torch.Tensor): Weights from the new run of the competition.
        """
        # Check that the weights are the appropriate length along the first dimension.
        if new_weights.size(0) != self.num_neurons:
            self._resize(new_weights.size(0))

        # Normalize the weights.
        new_weights /= new_weights.sum()
        new_weights = new_weights.nan_to_num(0.0)

        # If we haven't recorded weights for this competition yet, start from scratch.
        if competition_id not in self.weights_by_competition:
            self.weights_by_competition[competition_id] = new_weights
        # Otherwise compute a moving average using alpha.
        else:
            moving_average_weights = (
                self.alpha * self.weights_by_competition[competition_id]
                + (1 - self.alpha) * new_weights
            )
            self.weights_by_competition[competition_id] = moving_average_weights

    def reset_competitions(self, competition_ids: typing.Set[int]):
        """Resets tracked competitions to only those identified.

        Args:
            competition_ids (typing.Set[int]): Competition ids to continue tracking.
        """
        # Make a list to avoid issues deleting from within the dictionary iterator.
        for key in list(self.weights_by_competition.keys()):
            if key not in competition_ids:
                del self.weights_by_competition[key]

    def get_subnet_weights(
        self,
        competitions: typing.List[Competition],
        min_comp_weight_threshold: float = 0.0,
    ) -> torch.Tensor:
        """Gets the weights that should be set on the subnet level based on all competitions.

        Args:
            competitions (typing.List[Competition]): Competitions to calculate weights across.
            min_comp_weight_threshold (float): Competition weights below this threshold will be dropped.

        Returns:
            torch.Tensor: Weights calculated across all specified competitions.
        """
        # Return a copy to ensure outside code can't modify the scores.
        # Start each uid at 0.
        subnet_weights = torch.zeros(self.num_neurons, dtype=torch.float32)

        # For each competition, add the relative competition weight
        for competition in competitions:
            if competition.id in self.weights_by_competition:
                comp_weights = self.weights_by_competition[competition.id]
                # If we were provided a threshold under which to 0 out weights then do so.
                if min_comp_weight_threshold > 0:
                    # Normalize competition weight here to be safe although they should already be normalized.
                    comp_weights /= comp_weights.sum()
                    comp_weights = comp_weights.nan_to_num(0.0)
                    # Drop weights below the threshold.
                    comp_weights[comp_weights < min_comp_weight_threshold] = 0.0
                    # Normalize again because not all competitions will have any dropped weights.
                    comp_weights /= comp_weights.sum()
                    comp_weights = comp_weights.nan_to_num(0.0)
                for i in range(self.num_neurons):
                    # Today uids can only participate in one competition so += and = would be equivalent.
                    subnet_weights[i] += comp_weights[i] * competition.reward_percentage

        # Normalize weights again in case a competition hasn't run yet.
        # In the case that in Block X we have Competition 1 at 100%, but Block X+1 we have
        # Competition 1 at 80% and Competition 2 at 20%, but we haven't run Competition 2 yet.
        # This ensures we still return 1 for the winner of competition 1 instead of 0.8.
        subnet_weights /= subnet_weights.sum()
        subnet_weights = subnet_weights.nan_to_num(0.0)

        return subnet_weights

    def get_competition_weights(self, competition_id: int) -> torch.Tensor:
        """Returns the current weights for a single competition"""
        # Return a copy to ensure outside code can't modify the scores.
        return self.weights_by_competition[competition_id].clone()

    def _resize(self, new_num_neurons: int) -> None:
        """Resizes the score tensor to the new number of neurons.

        The new size must be greater than or equal to the current size.
        """
        assert (
            new_num_neurons >= self.num_neurons
        ), f"Tried to downsize the number of neurons from {self.num_neurons} to {new_num_neurons}"

        bt.logging.trace(
            f"Resizing CompetitionTracker from {self.num_neurons} to {new_num_neurons}"
        )

        # Compute additional number of neurons to add to the end of each competition weight tensor.
        to_add = new_num_neurons - self.num_neurons
        for competition in self.weights_by_competition:
            self.weights_by_competition[competition] = torch.cat(
                [
                    self.weights_by_competition[competition],
                    torch.zeros(to_add, dtype=torch.float32),
                ]
            )

        # Keep track of the new size of each competition weight tensor.
        self.num_neurons = new_num_neurons

    def save_state(self, filepath):
        """Save the current state to the provided filepath."""
        torch.save(
            {
                "weights_by_competition": self.weights_by_competition,
                "num_neurons": self.num_neurons,
            },
            filepath,
        )

    def load_state(self, filepath):
        """Load the state from the provided filepath."""
        state = torch.load(filepath)
        self.weights_by_competition = state["weights_by_competition"]
        self.num_neurons = state["num_neurons"]



import abc
from typing import Optional
from taoverse.model.data import ModelId, ModelMetadata


class ModelMetadataStore(abc.ABC):
    """An abstract base class for storing and retrieving model metadata."""

    @abc.abstractmethod
    async def store_model_metadata(
        self,
        hotkey: str,
        model_id: ModelId,
        wait_for_inclusion: bool = True,
        wait_for_finalization: bool = True,
    ):
        """Stores model metadata on this subnet for a specific miner."""
        pass

    @abc.abstractmethod
    async def retrieve_model_metadata(self, hotkey: str) -> Optional[ModelMetadata]:
        """Retrieves model metadata + block information on this subnet for specific miner, if present"""
        pass



import dataclasses

from taoverse.model.eval.normalization import NormalizationId

@dataclasses.dataclass
class EvalTask:
    """Represents a task to evaluate a model on."""

    # Friendly name for the task.
    name: str

    # The identifier of the evaluation method to use.
    method_id: int
    
    # The identifier of the dataset to evaluate on.
    dataset_id: int

    # The identifier of the normalization method to use.
    normalization_id: NormalizationId = NormalizationId.NONE

    # Additional keyword arguments to pass to the normalization method.
    normalization_kwargs: dict = dataclasses.field(default_factory=dict)
    
    # Additional keyword arguments to pass to the dataset loader.
    dataset_kwargs: dict = dataclasses.field(default_factory=dict)

    # Weight to apply to the normalized score to provide relative weight against other EvalTasks.
    weight: float = 1.0

    def __post_init__(self):
        if self.weight <= 0:
            raise ValueError("Weight must be positive.")

        match self.normalization_id:
            case NormalizationId.NONE:
                if self.normalization_kwargs:
                    raise ValueError(
                        "Normalization kwargs should not be provided for NONE normalization."
                    )
            case NormalizationId.INVERSE_EXPONENTIAL:
                if "ceiling" not in self.normalization_kwargs:
                    raise ValueError(
                        "Normalization kwargs must contain a 'ceiling' value."
                    )


import unittest

from taoverse.model.competition.epsilon import FixedEpsilon, LinearDecay


class TestEpsilonFuncs(unittest.TestCase):
    def test_fixed_decay_is_always_fixed(self):
        epsilon_decay = FixedEpsilon(0.005)

        model_block = 100
        current_block = [95, 100, 105]
        for block in current_block:
            self.assertEqual(
                epsilon_decay.compute_epsilon(
                    model_block=model_block, current_block=block
                ),
                0.005,
            )

    def test_linear_decay_current_block_earlier_than_model_block(self):
        epsilon_decay = LinearDecay(
            start_epsilon=0.005,
            end_epsilon=0.001,
            decay_blocks=7200 * 7,
        )

        calculated_epsilon = epsilon_decay.compute_epsilon(
            current_block=0, model_block=7200
        )

        self.assertEqual(calculated_epsilon, epsilon_decay.start_epsilon)

    def test_linear_decay_model_block_is_current_block(self):
        epsilon_decay = LinearDecay(
            start_epsilon=0.005,
            end_epsilon=0.001,
            decay_blocks=7200 * 7,
        )

        calculated_epsilon = epsilon_decay.compute_epsilon(
            current_block=7200, model_block=7200
        )

        self.assertEqual(calculated_epsilon, epsilon_decay.start_epsilon)

    def test_linear_decay_currnet_block_is_middle_block(self):
        epsilon_decay = LinearDecay(
            start_epsilon=0.005,
            end_epsilon=0.001,
            decay_blocks=7200 * 7,
        )

        calculated_epsilon = epsilon_decay.compute_epsilon(
            current_block=epsilon_decay.decay_blocks / 2,
            model_block=0,
        )

        self.assertEqual(
            calculated_epsilon,
            (epsilon_decay.start_epsilon + epsilon_decay.end_epsilon) / 2,
        )

    def test_linear_decay_current_block_is_decay_blocks(self):
        epsilon_decay = LinearDecay(
            start_epsilon=0.005,
            end_epsilon=0.001,
            decay_blocks=7200 * 7,
        )

        calculated_epsilon = epsilon_decay.compute_epsilon(
            current_block=7200 * 8,
            model_block=7200,
        )

        self.assertEqual(calculated_epsilon, epsilon_decay.end_epsilon)

    def test_linear_decay_current_block_beyond_max_block(self):
        epsilon_decay = LinearDecay(
            start_epsilon=0.005,
            end_epsilon=0.001,
            decay_blocks=7200 * 7,
        )

        calculated_epsilon = epsilon_decay.compute_epsilon(
            current_block=7200 * 100,
            model_block=7200,
        )

        self.assertEqual(calculated_epsilon, epsilon_decay.end_epsilon)


if __name__ == "__main__":
    unittest.main()



import unittest

from taoverse.metagraph.miner_iterator import MinerIterator


class TestMinerIterator(unittest.TestCase):
    def test_miner_uids_are_sorted(self):
        """Creates a MinerIterator with unsorted miner UIDs and verifies that the miner UIDs are sorted."""
        uids = [2, 5, 1, 0]
        iterator = MinerIterator(uids)

        # The iterator starts at a random position. Move it until we're pointing to 0.
        while iterator.peek() != 0:
            next(iterator)

        # Now verify the UIDs are iterated in sorted order.
        iterated_uids = [next(iterator) for _ in range(len(uids))]
        self.assertEqual(iterated_uids, sorted(uids))

    def test_iterator_is_infinite(self):
        """Creates a MinerIterator and verifies calling it more times than the number of miner UIDs cycles the UIDs."""
        uids = [3, 2, 1]
        expected = [1, 2, 3] * 10
        iterator = MinerIterator(uids)
        iterated_uids = [next(iterator) for _ in range(30)]
        self.assertEqual(sorted(iterated_uids), sorted(expected))

    def test_peek(self):
        """Creates a MinerIterator and verifies that peek returns the next UID without advancing the iterator."""
        uids = [1, 2, 3]
        iterator = MinerIterator(uids)

        peeked = iterator.peek()
        self.assertEqual(peeked, iterator.peek())
        self.assertEqual(peeked, next(iterator))
        self.assertNotEqual(peeked, iterator.peek())

    def test_set_miner_uids(self):
        """Verifies the iterator position is maintained when the miner UIDs are updated."""
        initial_miner_uids = [1, 2, 3, 4, 5]
        iterator = MinerIterator(initial_miner_uids)

        # Advance the iterator so it should now point to 3
        # The iterator starts at a random position. Advance it until it returns 2.
        while next(iterator) != 2:
            pass

        iterator.set_miner_uids([1, 4, 6])

        # Verify the iterator picks up from the next UID greater than or equal to 3.
        self.assertEqual(next(iterator), 4)
        self.assertEqual(next(iterator), 6)
        self.assertEqual(next(iterator), 1)

    def test_set_miner_uids_edge_case(self):
        """Verifies the iterator position is reset when the miner UIDs are updated and the current position is no longer valid."""
        # Create a MinerIterator with initial miner UIDs
        initial_miner_uids = [1, 2, 3, 4, 5]
        iterator = MinerIterator(initial_miner_uids)

        # Advance the iterator so it should now point to 5
        while iterator.peek() != 5:
            next(iterator)

        iterator.set_miner_uids([1, 2, 3, 4])

        self.assertEqual(next(iterator), 1)
        self.assertEqual(next(iterator), 2)
        self.assertEqual(next(iterator), 3)
        self.assertEqual(next(iterator), 4)


if __name__ == "__main__":
    unittest.main()



import unittest

import torch
from transformers import (
    BartForCausalLM,
    FalconForCausalLM,
    GemmaForCausalLM,
    GPTNeoXForCausalLM,
    LlamaForCausalLM,
    MistralForCausalLM,
    PhiForCausalLM,
)
from typing import Dict, List, Tuple

from taoverse.model.competition.data import (
    Competition,
    ModelConstraints,
    NormValidationConstraints,
)
from taoverse.model.competition.epsilon import FixedEpsilon
from taoverse.model.competition.utils import (
    get_competition_for_block,
    get_competition_schedule_for_block,
)

unittest.util._MAX_LENGTH = 2000


class TestCompetitionUtils(unittest.TestCase):
    COMPETITION_ID = 1

    MODEL_CONSTRAINTS_BY_COMPETITION_ID: Dict[int, ModelConstraints] = {
        COMPETITION_ID: ModelConstraints(
            max_model_parameter_size=6_900_000_000,
            sequence_length=4096,
            allowed_architectures=[
                MistralForCausalLM,
                LlamaForCausalLM,
                BartForCausalLM,
                FalconForCausalLM,
                GPTNeoXForCausalLM,
                PhiForCausalLM,
                GemmaForCausalLM,
            ],
            tokenizer="Xenova/gpt-4",
            kwargs={
                "torch_dtype": torch.bfloat16,
            },
            eval_block_delay=1200,  # ~4 hours.
            norm_validation_constraints=NormValidationConstraints(
                norm_eps_soft=200,
                norm_eps_soft_percent_threshold=0.15,
                norm_eps_hard=1000,
            ),
            epsilon_func=FixedEpsilon(0.005),
            max_bytes=15*1024*1024*1024,  # 15 GB
        ),
    }

    COMPETITION_SCHEDULE_BY_BLOCK: List[Tuple[int, List[Competition]]] = [
        (
            0,
            [
                Competition(
                    COMPETITION_ID,
                    MODEL_CONSTRAINTS_BY_COMPETITION_ID[COMPETITION_ID],
                    1.0,
                )
            ],
        )
    ]

    def test_get_competition_for_block_valid_competition(self):
        expected_competition = Competition(
            id=self.COMPETITION_ID,
            constraints=ModelConstraints(
                max_model_parameter_size=6_900_000_000,
                sequence_length=4096,
                allowed_architectures=[
                    MistralForCausalLM,
                    LlamaForCausalLM,
                    BartForCausalLM,
                    FalconForCausalLM,
                    GPTNeoXForCausalLM,
                    PhiForCausalLM,
                    GemmaForCausalLM,
                ],
                tokenizer="Xenova/gpt-4",
                eval_block_delay=1200,
                kwargs={
                    "torch_dtype": torch.bfloat16,
                },
                norm_validation_constraints=NormValidationConstraints(
                    norm_eps_soft=200,
                    norm_eps_soft_percent_threshold=0.15,
                    norm_eps_hard=1000,
                ),
                epsilon_func=FixedEpsilon(0.005),
                max_bytes=15*1024*1024*1024,  # 15 GB
            ),
            reward_percentage=1.0,
        )

        competition = get_competition_for_block(
            comp_id=self.COMPETITION_ID,
            block=0,
            schedule_by_block=self.COMPETITION_SCHEDULE_BY_BLOCK,
        )
        self.assertEqual(competition, expected_competition)

    def test_get_competition_for_block_invalid_competition(self):
        competition = get_competition_for_block(
            comp_id=-1, block=0, schedule_by_block=self.COMPETITION_SCHEDULE_BY_BLOCK
        )
        self.assertIsNone(competition)

    def test_get_competition_for_block_invalid_block(self):
        with self.assertRaises(Exception):
            _ = get_competition_for_block(
                comp_id=self.COMPETITION_ID,
                block=-1,
                schedule_by_block=self.COMPETITION_SCHEDULE_BY_BLOCK,
            )

    def test_get_competition_schedule_for_block_valid_block(self):
        expected_competition_schedule = [
            Competition(
                id=self.COMPETITION_ID,
                constraints=ModelConstraints(
                    max_model_parameter_size=6_900_000_000,
                    sequence_length=4096,
                    allowed_architectures=[
                        MistralForCausalLM,
                        LlamaForCausalLM,
                        BartForCausalLM,
                        FalconForCausalLM,
                        GPTNeoXForCausalLM,
                        PhiForCausalLM,
                        GemmaForCausalLM,
                    ],
                    tokenizer="Xenova/gpt-4",
                    eval_block_delay=1200,
                    kwargs={
                        "torch_dtype": torch.bfloat16,
                    },
                    norm_validation_constraints=NormValidationConstraints(
                        norm_eps_soft=200,
                        norm_eps_soft_percent_threshold=0.15,
                        norm_eps_hard=1000,
                    ),
                    epsilon_func=FixedEpsilon(0.005),
                    max_bytes=15*1024*1024*1024,  # 15 GB
                ),
                reward_percentage=1.0,
            ),
        ]

        competition_schedule = get_competition_schedule_for_block(
            block=0, schedule_by_block=self.COMPETITION_SCHEDULE_BY_BLOCK
        )
        self.assertEqual(competition_schedule, expected_competition_schedule)

    def test_get_competition_schedule_for_block_invalid_block(self):
        with self.assertRaises(Exception):
            _ = get_competition_schedule_for_block(
                block=-1, schedule_by_block=self.COMPETITION_SCHEDULE_BY_BLOCK
            )


if __name__ == "__main__":
    unittest.main()
